@@ -14,9 +14,9 @@ import {ViewUpdate, styleModule,
         contentAttributes, editorAttributes, AttrSource,
         clickAddsSelectionRange, dragMovesSelection, mouseSelectionStyle,
         exceptionSink, updateListener, logException,
-        viewPlugin, ViewPlugin, PluginValue, PluginInstance, decorations, atomicRanges,
+        viewPlugin, ViewPlugin, PluginValue, PluginInstance, decorations, outerDecorations, atomicRanges,
         scrollMargins, MeasureRequest, editable, inputHandler, focusChangeEffect, perLineTextDirection,
-        scrollIntoView, UpdateFlag, ScrollTarget, bidiIsolatedRanges, getIsolatedRanges} from "./extension"
+        scrollIntoView, UpdateFlag, ScrollTarget, bidiIsolatedRanges, getIsolatedRanges, scrollHandler} from "./extension"
 import {theme, darkTheme, buildTheme, baseThemeID, baseLightID, baseDarkID, lightDarkIDs, baseTheme} from "./theme"
 import {DOMObserver} from "./domobserver"
 import {Attrs, updateAttrs, combineAttrs} from "./attributes"
@@ -322,6 +322,7 @@ export class EditorView {
       this.viewState.mustMeasureContent = true
     if (redrawn || attrsChanged || scrollTarget || this.viewState.mustEnforceCursorAssoc || this.viewState.mustMeasureContent)
       this.requestMeasure()
+    if (redrawn) this.docViewUpdate()
     if (!update.empty) for (let listener of this.state.facet(updateListener)) {
       try { listener(update) }
       catch (e) { logException(this.state, e, "update listener") }
@@ -391,6 +392,16 @@ export class EditorView {
     if (prevSpecs != specs) this.inputState.ensureHandlers(this.plugins)
   }
 
+  private docViewUpdate() {
+    for (let plugin of this.plugins) {
+      let val = plugin.value
+      if (val && val.docViewUpdate) {
+        try { val.docViewUpdate(this) }
+        catch(e) { logException(this.state, e, "doc view update listener") }
+      }
+    }
+  }
+
   /// @internal
   measure(flush = true) {
     if (this.destroyed) return
@@ -449,6 +460,7 @@ export class EditorView {
           this.inputState.update(update)
           this.updateAttrs()
           redrawn = this.docView.update(update)
+          if (redrawn) this.docViewUpdate()
         }
         for (let i = 0; i < measuring.length; i++) if (measured[i] != BadMeasure) {
           try {
@@ -661,6 +673,16 @@ export class EditorView {
     return skipAtoms(this, start, moveByChar(this, start, forward, initial => byGroup(this, start.head, initial)))
   }
 
+  /// Get the cursor position visually at the start or end of a line.
+  /// Note that this may differ from the _logical_ position at its
+  /// start or end (which is simply at `line.from`/`line.to`) if text
+  /// at the start or end goes against the line's base text direction.
+  visualLineSide(line: Line, end: boolean) {
+    let order = this.bidiSpans(line), dir = this.textDirectionAt(line.from)
+    let span = order[end ? order.length - 1 : 0]
+    return EditorSelection.cursor(span.side(end, dir) + line.from, span.forward(!end, dir) ? 1 : -1)
+  }
+
   /// Move to the next line boundary in the given direction. If
   /// `includeWrap` is true, line wrapping is on, and there is a
   /// further wrap point on the current line, the wrap point will be
@@ -785,10 +807,10 @@ export class EditorView {
     let dir = this.textDirectionAt(line.from), isolates: readonly Isolate[] | undefined
     for (let entry of this.bidiCache) {
       if (entry.from == line.from && entry.dir == dir &&
-          (entry.fresh || isolatesEq(entry.isolates, isolates = getIsolatedRanges(this, line.from, line.to))))
+          (entry.fresh || isolatesEq(entry.isolates, isolates = getIsolatedRanges(this, line))))
         return entry.order
     }
-    if (!isolates) isolates = getIsolatedRanges(this, line.from, line.to)
+    if (!isolates) isolates = getIsolatedRanges(this, line)
     let order = computeOrder(line.text, dir, isolates)
     this.bidiCache.push(new CachedOrder(line.from, line.to, dir, isolates, true, order))
     return order
@@ -923,6 +945,12 @@ export class EditorView {
   /// dispatching the custom behavior as a separate transaction.
   static inputHandler = inputHandler
 
+  /// Scroll handlers can override how things are scrolled into view.
+  /// If they return `true`, no further handling happens for the
+  /// scrolling. If they return false, the default scroll behavior is
+  /// applied. Scroll handlers should never initiate editor updates.
+  static scrollHandler = scrollHandler
+
   /// This facet can be used to provide functions that create effects
   /// to be dispatched when the editor's focus state changes.
   static focusChangeEffect = focusChangeEffect
@@ -986,6 +1014,15 @@ export class EditorView {
   /// [`EditorView.atomicRanges`](#view.EditorView^atomicRanges).
   static decorations = decorations
 
+  /// Facet that works much like
+  /// [`decorations`](#view.EditorView^decorations), but puts its
+  /// inputs at the very bottom of the precedence stack, meaning mark
+  /// decorations provided here will only be split by other, partially
+  /// overlapping \`outerDecorations\` ranges, and wrap around all
+  /// regular decorations. Use this for mark elements that should, as
+  /// much as possible, remain in one piece.
+  static outerDecorations = outerDecorations
+
   /// Used to provide ranges that should be treated as atoms as far as
   /// cursor motion is concerned. This causes methods like
   /// [`moveByChar`](#view.EditorView.moveByChar) and
