@@ -1,4 +1,4 @@
-import {EditorSelection, EditorState, SelectionRange, RangeSet, Annotation} from "@codemirror/state"
+import {EditorSelection, EditorState, SelectionRange, RangeSet, Annotation, Text} from "@codemirror/state"
 import {EditorView} from "./editorview"
 import {ContentView} from "./contentview"
 import {LineView} from "./blockview"
@@ -20,7 +20,7 @@ export class InputState {
   // On iOS, some keys need to have their default behavior happen
   // (after which we retroactively handle them and reset the DOM) to
   // avoid messing up the virtual keyboard state.
-  pendingIOSKey: undefined | {key: string, keyCode: number} = undefined
+  pendingIOSKey: undefined | {key: string, keyCode: number} | KeyboardEvent = undefined
 
   lastSelectionOrigin: string | null = null
   lastSelectionTime: number = 0
@@ -142,11 +142,13 @@ export class InputState {
     return false
   }
 
-  flushIOSKey() {
+  flushIOSKey(change?: {from: number, to: number, insert: Text}) {
     let key = this.pendingIOSKey
     if (!key) return false
+    // This looks like an autocorrection before Enter
+    if (key.key == "Enter" && change && change.from < change.to && /^\S+$/.test(change.insert.toString())) return false
     this.pendingIOSKey = undefined
-    return dispatchKey(this.view.contentDOM, key.key, key.keyCode)
+    return dispatchKey(this.view.contentDOM, key.key, key.keyCode, key instanceof KeyboardEvent ? key : undefined)
   }
 
   ignoreDuringComposition(event: Event): boolean {
@@ -381,8 +383,7 @@ class MouseSelection {
 
   select(event: MouseEvent) {
     let {view} = this, selection = this.skipAtoms(this.style.get(event, this.extend, this.multiple))
-    if (this.mustSelect || !selection.eq(view.state.selection) ||
-        selection.main.assoc != view.state.selection.main.assoc && this.dragging === false)
+    if (this.mustSelect || !selection.eq(view.state.selection, this.dragging === false))
       this.view.dispatch({
         selection,
         userEvent: "select.pointer"
@@ -697,7 +698,7 @@ handlers.paste = (view: EditorView, event: ClipboardEvent) => {
   view.observer.flush()
   let data = brokenClipboardAPI ? null : event.clipboardData
   if (data) {
-    doPaste(view, data.getData("text/plain") || data.getData("text/uri-text"))
+    doPaste(view, data.getData("text/plain") || data.getData("text/uri-list"))
     return true
   } else {
     capturePaste(view)
@@ -866,6 +867,17 @@ handlers.beforeinput = (view, event) => {
       }, 100)
     }
   }
+  if (browser.ios && event.inputType == "deleteContentForward") {
+    // For some reason, DOM changes (and beforeinput) happen _before_
+    // the key event for ctrl-d on iOS when using an external
+    // keyboard.
+    view.observer.flushSoon()
+  }
+  // Safari will occasionally forget to fire compositionend at the end of a dead-key composition
+  if (browser.safari && event.inputType == "insertText" && view.inputState.composing >= 0) {
+    setTimeout(() => observers.compositionend(view, event), 20)
+  }
+
   return false
 }
 
