@@ -114,7 +114,7 @@ export class ViewState {
   editorHeight = 0 // scrollDOM.clientHeight, unscaled
   editorWidth = 0 // scrollDOM.clientWidth, unscaled
   scrollTop = 0 // Last seen scrollDOM.scrollTop, scaled
-  scrolledToBottom = true
+  scrolledToBottom = false
   // The CSS-transformation scale of the editor (transformed size /
   // concrete size)
   scaleX = 1
@@ -256,7 +256,8 @@ export class ViewState {
 
     if (domRect.width && domRect.height) {
       let {scaleX, scaleY} = getScale(dom, domRect)
-      if (this.scaleX != scaleX || this.scaleY != scaleY) {
+      if (scaleX > .005 && Math.abs(this.scaleX - scaleX) > .005 ||
+          scaleY > .005 && Math.abs(this.scaleY - scaleY) > .005) {
         this.scaleX = scaleX; this.scaleY = scaleY
         result |= UpdateFlag.Geometry
         refresh = measureContent = true
